/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_SYNCHRONIZER_HXX_
# define WIFI_FRAME_FILTER_SYNCHRONIZER_HXX_

# include <trace-tools/tool/linear_regression.hh>

# include "linear_regression_synchronizer.hh"

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      template <class OriginalFrameType, class ImplType>
      synchronized_frame<OriginalFrameType, ImplType>::
      synchronized_frame(const OriginalFrameType&	orig,
			 const coefs_type&		coefs):
	OriginalFrameType (orig),
	coefs (coefs)
      {
      }

      namespace internals
      {

	template <class I, class B1, class B2>
	lr_sync_iterator<I, B1, B2>::
	lr_sync_iterator(linear_regression_synchronizer<I, B1>& i, bool end):
	  i_ (&i),
	  next_wpos_ (0),
	  v_ ()
	{
	  if (not end)
	    increment();
	}

	template <class I, class B1, class B2>
	bool
	lr_sync_iterator<I, B1, B2>::equal(const exact_type& rhs) const
	{
	  return
	    i_ == rhs.i_ and next_wpos_ == rhs.next_wpos_ and v_ == rhs.v_;
	}

	template <class I, class B1, class B2>
	void
	lr_sync_iterator<I, B1, B2>::increment()
	{
	  typedef typename value_type::impl_type	impl_type;

	  typedef
	    typename tool::lr::pair_with_microseconds<impl_type>
	    adapter_type;

	  const I&	last = i_->last_;
	  I&		next = i_->next_;
	  const size_t	window_size = next->size();

	  if (next_wpos_ == window_size and next == last)
	    {
	      next_wpos_ = 0;
	      v_ = boost::none_t ();
	      return;
	    }

	  v_ =
	    value_type (next[next_wpos_],
			tool::linear_regression<impl_type,
						adapter_type> (next->begin(),
							       next->end()));

	  if (next_wpos_ < window_size / 2 or next == last)
	    ++next_wpos_;
	  else
	    ++next;
	}

	template <class I, class B1, class B2>
	const typename lr_sync_iterator<I, B1, B2>::value_type&
	lr_sync_iterator<I, B1, B2>::get() const
	{
	  return v_.get();
	}

	template <class I, class B1, class B2>
	const typename lr_sync_iterator<I, B1, B2>::value_type*
	lr_sync_iterator<I, B1, B2>::get_ptr() const
	{
	  return v_.get_ptr();
	}

      } // End of namespace wifi::frame::filter::internals.


      template <class I, class B>
      linear_regression_synchronizer<I, B>::
      linear_regression_synchronizer(const I& first, const I& last):
	next_ (first),
	last_ (last)
      {
      }

    } // End of namespace wifi::frame::filter.
  } // End of namespace wifi::frame.
} // End of namespace wifi.

#endif // ! WIFI_FRAME_FILTER_SYNCHRONIZER_HXX_
