/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#include <iostream>
#include <stdexcept>
#include <cstdlib>
#include <unistd.h>

#include "config.hh"

#include <trace-tools/pcap/descriptor.hh>
#include <trace-tools/wifi/addr_mapping.hh>
#include <trace-tools/wifi/frame/unique_id/seqctl_dest_bssid_timestamp.hh>
#include <trace-tools/wifi/frame/filter/uniquely_identifiable.hh>
#include <trace-tools/wifi/frame/filter/intersector.hh>

#define ADDR_MAPPING_DEFAULT	"mapping"

namespace
{

  struct flags
  {
    bool	prism_header_noise_means_error;
    bool	print_microseconds;
  };

  void
  usage(int, char* const argv[])
  {
    std::cerr << "Usage:\n"
      "\t" << argv[0] << " [options] <trace1> <trace2>\n"
      "\nOptions:\n"
      "\t-h\tPrint this help.\n"
      "\t-m\tAddress mapping file (default: " ADDR_MAPPING_DEFAULT ").\n"
      "\t-p\tPrism noise fields have no specific signification (default).\n"
      "\t-P\tIgnore frames with non-zero Prism noise fields.\n"
      "\t-u\tDo not print microseconds (default).\n"
      "\t-U\tPrint microseconds.\n";
  }

  void
  parse_args(int argc, char* const argv[], std::string& map, flags& f)
  {
    int opt;

    while ((opt = getopt(argc, argv, "m:hpPuU")) != -1)
      switch (opt)
	{
	case 'h':
	  usage(argc, argv);
	  exit(EXIT_SUCCESS);

	case 'm':
	  map = optarg;
	  break;

	case 'p':
	  f.prism_header_noise_means_error = false;
	  break;

	case 'P':
	  f.prism_header_noise_means_error = true;
	  break;

	case 'u':
	  f.print_microseconds = false;
	  break;

	case 'U':
	  f.print_microseconds = true;
	  break;

	default: /* '?' */
	  usage(argc, argv);
	  exit(EXIT_FAILURE);
	}
    if (optind != argc - 2)
    {
      std::cerr << "ERROR: You must specify two traces to intersect."
		<< std::endl;
      exit(EXIT_FAILURE);
    }
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  std::string	map;
  flags		f;

  map = ADDR_MAPPING_DEFAULT;
  f.prism_header_noise_means_error = false;
  f.print_microseconds = false;
  parse_args(argc, argv, map, f);

  try
    {
      typedef wifi::frame::filter::
	uniquely_identifiable<wifi::frame::seqctl_dest_bssid_timestamp_id,
			      pcapxx::descriptor<>::iterator>
	uniquely_identifiable;
      typedef
	wifi::frame::filter::intersector<uniquely_identifiable::iterator,
					 uniquely_identifiable::iterator>
	intersector;

      wifi::addr_mapping	m (map);
      pcapxx::descriptor<>	d1 (argv[optind]);
      pcapxx::descriptor<>	d2 (argv[optind + 1]);
      uniquely_identifiable	u1 (d1.begin(), d1.end(), m);
      uniquely_identifiable	u2 (d2.begin(), d2.end(), m);

      u1.prism_header_noise_means_error = f.prism_header_noise_means_error;
      u2.prism_header_noise_means_error = f.prism_header_noise_means_error;

      intersector		i (u1.begin(), u1.end(), u2.begin(), u2.end());

      for (intersector::iterator j = i.begin(); j != i.end(); ++j)
	{
	  if (f.print_microseconds)
	    std::cout << j->first.microseconds() << '\t';
	  std::cout << j->first << '\t';
	  if (f.print_microseconds)
	    std::cout << j->second.microseconds() << '\t';
	  std::cout << j->second << '\n';
	}
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
