/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#include <iostream>
#include <stdexcept>
#include <cstdlib>
#include <gmpxx.h>
#include <unistd.h>

#include <boost/format.hpp>

#include "config.hh"

#include <trace-tools/pcap/descriptor.hh>
#include <trace-tools/wifi/addr_mapping.hh>
#include <trace-tools/wifi/frame/unique_id/seqctl_dest_bssid_timestamp.hh>
#include <trace-tools/wifi/frame/filter/uniquely_identifiable.hh>
#include <trace-tools/wifi/frame/filter/intersector.hh>
#include <trace-tools/tool/window.hh>
#include <trace-tools/wifi/frame/filter/linear_regression_synchronizer.hh>

#define ADDR_MAPPING_DEFAULT	"mapping"

namespace
{

  struct flags
  {
    bool	print_column_headers;
    bool	print_errors;
    bool	print_frame_indices;
    bool	print_microseconds;
    bool	prism_header_noise_means_error;
  };

  void
  usage(int, char* const argv[])
  {
    std::cerr << "Usage:\n"
      "\t" << argv[0] << " [options] <trace1> <trace2>\n"
      "\nOptions:\n"
      "\t-c\tDo not print column headers (default if output is not a tty).\n"
      "\t-C\tPrint column headers (default if output is a tty).\n"
      "\t-e\tDo not print error values (default).\n"
      "\t-E\tPrint error values.\n"
      "\t-h\tPrint this help.\n"
      "\t-i\tDo not print frame indices.\n"
      "\t-I\tPrint frame indices (default).\n"
      "\t-m\tAddress mapping file (default: " ADDR_MAPPING_DEFAULT ").\n"
      "\t-p\tPrism noise fields have no specific signification (default).\n"
      "\t-P\tIgnore frames with non-zero Prism noise fields.\n"
      "\t-u\tDo not print microseconds (default).\n"
      "\t-U\tPrint microseconds.\n";
  }

  void
  parse_args(int argc, char* const argv[], std::string& map, flags& f)
  {
    int opt;

    while ((opt = getopt(argc, argv, "m:cCeEhiIpPuU")) != -1)
      switch (opt)
	{
	case 'c':
	  f.print_column_headers = false;
	  break;

	case 'C':
	  f.print_column_headers = true;
	  break;

	case 'e':
	  f.print_errors = false;
	  break;

	case 'E':
	  f.print_errors = true;
	  break;

	case 'h':
	  usage(argc, argv);
	  exit(EXIT_SUCCESS);

	case 'i':
	  f.print_frame_indices = false;
	  break;

	case 'I':
	  f.print_frame_indices = true;
	  break;

	case 'm':
	  map = optarg;
	  break;

	case 'p':
	  f.prism_header_noise_means_error = false;
	  break;

	case 'P':
	  f.prism_header_noise_means_error = true;
	  break;

	case 'u':
	  f.print_microseconds = false;
	  break;

	case 'U':
	  f.print_microseconds = true;
	  break;

	default: /* '?' */
	  usage(argc, argv);
	  exit(EXIT_FAILURE);
	}
    if (optind != argc - 2)
    {
      std::cerr << "ERROR: You must specify two traces to synchronize."
		<< std::endl;
      exit(EXIT_FAILURE);
    }
  }

  void
  print_column_headers(boost::format&	coefs,
		       boost::format&	indices,
		       boost::format&	msecs,
		       boost::format&	errors,
		       const flags&	f)
  {
    std::cout << (coefs % "a" % "b");
    if (f.print_frame_indices)
      std::cout << (indices % "frame 1" % "frame 2");
    if (f.print_microseconds)
      std::cout << (msecs % "time 1 (us)" % "time 2 (us)");
    if (f.print_errors)
      std::cout << (errors % "error");

    std::cout << '\n'
	      << (coefs % "=" % "=");
    if (f.print_frame_indices)
      std::cout << (indices % "=======" % "=======");
    if (f.print_microseconds)
      std::cout << (msecs % "===========" % "===========");
    if (f.print_errors)
      std::cout << (errors % "=====");
    std::cout << std::endl;
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  std::string	map;
  flags		f;

  map = ADDR_MAPPING_DEFAULT;
  f.print_column_headers = isatty(STDOUT_FILENO);
  f.print_errors = false;
  f.print_frame_indices = true;
  f.print_microseconds = false;
  f.prism_header_noise_means_error = false;
  parse_args(argc, argv, map, f);

  mpf_set_default_prec(80);

  try
    {
      typedef wifi::frame::filter::
	uniquely_identifiable<wifi::frame::seqctl_dest_bssid_timestamp_id,
			      pcapxx::descriptor<>::iterator>
	uniquely_identifiable;

      typedef wifi::frame::filter::
	intersector<uniquely_identifiable::iterator,
		    uniquely_identifiable::iterator>
	intersector;

      typedef
	tool::window_maker<intersector::iterator, 3>
	window_maker;

      typedef wifi::frame::filter::
	linear_regression_synchronizer<window_maker::iterator>
	synchronizer;

      wifi::addr_mapping	m (map);
      pcapxx::descriptor<>	d1 (argv[optind]);
      pcapxx::descriptor<>	d2 (argv[optind + 1]);
      uniquely_identifiable	u1 (d1.begin(), d1.end(), m);
      uniquely_identifiable	u2 (d2.begin(), d2.end(), m);

      u1.prism_header_noise_means_error = f.prism_header_noise_means_error;
      u2.prism_header_noise_means_error = f.prism_header_noise_means_error;

      intersector		i (u1.begin(), u1.end(), u2.begin(), u2.end());
      window_maker		w (i.begin(), i.end());
      synchronizer		s (w.begin(), w.end());

      boost::format		coefs   ("%10s %15s");
      boost::format		indices (" %10s %10s");
      boost::format		msecs   (" %15s %15s");
      boost::format		errors  (" %10s");

      if (f.print_column_headers)
	print_column_headers(coefs, indices, msecs, errors, f);

      for (synchronizer::iterator j = s.begin(); j != s.end(); ++j)
	{
	  std::cout << (coefs % j->coefs.first % j->coefs.second);
	  if (f.print_frame_indices)
	    std::cout << (indices % j->first % j->second);
	  if (f.print_microseconds)
	    std::cout << (msecs % j->first.microseconds()
			        % j->second.microseconds());
	  if (f.print_errors)
	    std::cout << (errors % (j->first.microseconds() * j->coefs.first +
				    j->coefs.second -
				    j->second.microseconds()));
	  std::cout << '\n';
	}
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
