/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#include <iostream>
#include <set>
#include <map>
#include <stdexcept>
#include <cstdlib>
#include <unistd.h>

#include <boost/format.hpp>

#include "config.hh"

#include <trace-tools/pcap/descriptor.hh>
#include <trace-tools/wifi/addr_mapping.hh>
#include <trace-tools/wifi/frame/unique_id/addr_timestamp.hh>
#include <trace-tools/wifi/frame/unique_id/source_bssid_timestamp.hh>
#include <trace-tools/wifi/frame/unique_id/seqctl_bssid_timestamp.hh>
#include <trace-tools/wifi/frame/unique_id/seqctl_dest_bssid_timestamp.hh>
#include <trace-tools/wifi/frame/filter/uniquely_identifiable.hh>

#define IN_DEFAULT		"-"
#define ADDR_MAPPING_DEFAULT	"mapping"

#define DST_TMP_OPT		"dest-timestamp"
#define SRC_TMP_OPT		"source-timestamp"
#define BSS_TMP_OPT		"bssid-timestamp"
#define SRC_BSS_TMP_OPT		"source-bssid-timestamp"
#define SEQ_BSS_TMP_OPT		"seqctl-bssid-timestamp"
#define SEQ_DST_BSS_TMP_OPT	"seqctl-dest-bssid-timestamp"

namespace
{

  struct flags
  {
    bool	prism_header_noise_means_error;
    bool	print_column_headers;
    bool	print_summary;
  };

  struct collision
  {
    collision(const unsigned rhs_id, const tool::microseconds& interval):
      rhs_id (rhs_id),
      interval (interval)
    {
    }

    unsigned		rhs_id;
    tool::microseconds	interval;
  };

  template <class Filter, class InputIterator>
  void
  test_uniqueness(const InputIterator&	begin,
		  const InputIterator&	end,
		  wifi::addr_mapping&	addr_mapping,
		  const flags&		fgs)
  {
    typedef typename Filter::iterator			filter_iterator;
    typedef typename filter_iterator::value_type	arrival;
    typedef std::set<arrival>				set_of_arrivals;
    typedef std::vector<collision>			collisions_vector;
    typedef std::map<arrival, collisions_vector>	collisions_map;
    typedef typename collisions_vector::const_iterator	cvec_iterator;
    typedef typename collisions_map::const_iterator	cmap_iterator;

    Filter		filter (begin, end, addr_mapping);
    set_of_arrivals	arrivals;
    collisions_map	collisions;
    unsigned		arrival_count = 0;
    unsigned		collision_count = 0;

    filter.prism_header_noise_means_error = fgs.prism_header_noise_means_error;

    for (filter_iterator i = filter.begin(); i != filter.end(); ++i)
      {
	typedef typename set_of_arrivals::iterator iterator;

	const std::pair<iterator, bool> r = arrivals.insert(*i);

	++arrival_count;
	if (not r.second)
	  {
	    const tool::microseconds dt (i->microseconds() -
					 r.first->microseconds());

	    ++collision_count;
	    collisions[*r.first].push_back(collision (i->frame_id(), dt));
	  }
      }

    boost::format f ("%15s %15s %15s");

    if (fgs.print_column_headers)
      {
	std::cout << (f % "frame 1" % "frame 2" % "interval (us)") << '\n';
	std::cout << (f % "=======" % "=======" % "=============") << std::endl;
      }
    for (cmap_iterator i = collisions.begin(); i != collisions.end(); ++i)
      {
	bool first = true;
	for (cvec_iterator j = i->second.begin(); j != i->second.end(); ++j)
	  {
	    std::cout << (f % i->first % j->rhs_id % j->interval);
	    if (first)
	      {
		first = false;
		std::cout << " (new)";
	      }
	    std::cout << '\n';
	  }
      }
    if (fgs.print_summary)
      std::cerr << arrival_count << " unique frames, "
		<< collision_count << " collisions." << std::endl;
  }

  void
  usage(int, char* const argv[])
  {
    std::cerr << "Usage:\n"
      "\t" << argv[0] << " [options] <distinctive_attributes>\n"
      "\nAttributes:\n"
      "\t" DST_TMP_OPT "\n"
      "\t" SRC_TMP_OPT "\n"
      "\t" BSS_TMP_OPT "\n"
      "\t" SRC_BSS_TMP_OPT "\n"
      "\t" SEQ_BSS_TMP_OPT "\n"
      "\t" SEQ_DST_BSS_TMP_OPT "\n"
      "\nOptions:\n"
      "\t-c\tDo not print column headers (default if  output is not a tty).\n"
      "\t-C\tPrint column headers (default if output is a tty).\n"
      "\t-h\tPrint this help.\n"
      "\t-i\tInput file (default: standard input).\n"
      "\t-m\tAddress mapping file (default: " ADDR_MAPPING_DEFAULT ").\n"
      "\t-p\tPrism noise fields have no specific signification (default).\n"
      "\t-P\tIgnore frames with non-zero Prism noise fields.\n"
      "\t-q\tQuiet; only print raw results (no column headers, no summary).\n";
  }

  void
  parse_args(int argc, char* const argv[],
	     std::string& in, std::string& map, flags& f)
  {
    int opt;

    while ((opt = getopt(argc, argv, "i:m:cChpPq")) != -1)
      switch (opt)
	{
	case 'c':
	  f.print_column_headers = false;
	  break;

	case 'C':
	  f.print_column_headers = true;
	  break;

	case 'h':
	  usage(argc, argv);
	  exit(EXIT_SUCCESS);

	case 'i':
	  in = optarg;
	  break;

	case 'm':
	  map = optarg;
	  break;

	case 'p':
	  f.prism_header_noise_means_error = false;
	  break;

	case 'P':
	  f.prism_header_noise_means_error = true;
	  break;

	case 'q':
	  f.print_column_headers = false;
	  f.print_summary = false;
	  break;

	default: /* '?' */
	  usage(argc, argv);
	  exit(EXIT_FAILURE);
	}

    if (optind != argc - 1)
      {
	usage(argc, argv);
	std::cerr << "ERROR: No distinctive attributes." << std::endl;
	exit(EXIT_FAILURE);
      }
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  std::string	in, map;
  flags		f;

  in = IN_DEFAULT;
  map = ADDR_MAPPING_DEFAULT;
  f.prism_header_noise_means_error = false;
  f.print_column_headers = isatty(STDOUT_FILENO);
  f.print_summary = true;
  parse_args(argc, argv, in, map, f);

  try
    {
      using namespace wifi::frame;

      typedef pcapxx::descriptor<>	pcap_descriptor;

#define tt_uifilter_typedef(UniqueId, Name)				\
      typedef filter::uniquely_identifiable<UniqueId,			\
					    pcap_descriptor::iterator> Name

      tt_uifilter_typedef(dest_timestamp_id,		  dst_tmp_f);
      tt_uifilter_typedef(source_timestamp_id,		  src_tmp_f);
      tt_uifilter_typedef(bssid_timestamp_id,		  bss_tmp_f);
      tt_uifilter_typedef(source_bssid_timestamp_id,	  src_bss_tmp_f);
      tt_uifilter_typedef(seqctl_bssid_timestamp_id,	  seq_bss_tmp_f);
      tt_uifilter_typedef(seqctl_dest_bssid_timestamp_id, seq_dst_bss_tmp_f);

#undef tt_uifilter_typedef

      wifi::addr_mapping	m (map);
      pcapxx::descriptor<>	d (in);
      const std::string		a (argv[optind]);

      if (a == DST_TMP_OPT)
	test_uniqueness<dst_tmp_f>(d.begin(), d.end(), m, f);
      else if (a == SRC_TMP_OPT)
	test_uniqueness<src_tmp_f>(d.begin(), d.end(), m, f);
      else if (a == BSS_TMP_OPT)
	test_uniqueness<bss_tmp_f>(d.begin(), d.end(), m, f);
      else if (a == SRC_BSS_TMP_OPT)
	test_uniqueness<src_bss_tmp_f>(d.begin(), d.end(), m, f);
      else if (a == SEQ_BSS_TMP_OPT)
	test_uniqueness<seq_bss_tmp_f>(d.begin(), d.end(), m, f);
      else if (a == SEQ_DST_BSS_TMP_OPT)
	test_uniqueness<seq_dst_bss_tmp_f>(d.begin(), d.end(), m, f);
      else
	throw std::invalid_argument ("Invalid distinctive attributes");
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
